# argparsers.py - Argument parser classes.
# Copyright (C) 2025 David Gasaway
# https://github.com/dgasaway/towebm

# This program is free software; you can redistribute it and/or modify it under the terms of the GNU
# General Public License as published by the Free Software Foundation; either version 2 of the
# License, or (at your option) any later version.
#
# This program is distributed in the hope that it will be useful, but WITHOUT ANY WARRANTY; without
# even the implied warranty of MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE. See the GNU
# General Public License for more details.
#
# You should have received a copy of the GNU General Public License along with this program; if not,
# see <http://www.gnu.org/licenses>.
from __future__ import annotations

import os
import re
import sys
from argparse import Action, ArgumentError, ArgumentParser, Namespace, _ArgumentGroup
from dataclasses import dataclass, field
from typing import Any, Sequence, TYPE_CHECKING
if TYPE_CHECKING:
    from towebm.formats import Container, AudioFormat, VideoFormat, MappedCodecArg

from towebm._version import __version__

# --------------------------------------------------------------------------------------------------
@dataclass
class Segment:
    """
    Represents a segment of the input file bound by ffmpeg duration strings.
    """
    start_str: str | None
    """
    An ffmpeg duration string specifying the start of the segment.
    """
    end_str: str | None
    """
    An ffmpeg duration string specifying the end of the segment.
    """
    duration_str: str | None
    """
    An ffmpeg duration string specifying the duration of the segment.
    """
    start: float = field(init=False)
    """
    The start position of the segment in seconds.
    """
    end: float | None = field(init=False)
    """
    The end position of the segment in seconds.
    """
    duration: float | None = field(init=False)
    """
    The duration of the segment in seconds.
    """
    def __post_init__(self):
        """
        Parse the `start_str`, `end_str`, and `duration_str` values to `start`, `end`, and
        `duration`, respectively.
        """
        self.start = 0.0 if self.start_str is None else Segment.duration_to_seconds(self.start_str)
        if self.end_str is not None:
            self.end = Segment.duration_to_seconds(self.end_str)
            self.duration = self.end - self.start
        elif self.duration_str is not None:
            self.duration = Segment.duration_to_seconds(self.duration_str)
            self.end = self.start + self.duration
        else:
            self.duration = None
            self.end = None

    # ----------------------------------------------------------------------------------------------
    @staticmethod
    def duration_to_seconds(duration: str) -> float | None:
        """
        Convert the specified ffmpeg duration string into a decimal representing the number of
        seconds represented by the duration string; None if the string is not parsable.
        """
        pattern = r'^((((?P<hms_grp1>\d*):)?((?P<hms_grp2>\d*):)?((?P<hms_secs>\d+([.]\d*)?)))|' \
                r'((?P<smu_value>\d+([.]\d*)?)(?P<smu_units>s|ms|us)))$'
        match = re.match(pattern, duration)
        if match:
            groups = match.groupdict()
            if groups['hms_secs'] is not None:
                value = float(groups['hms_secs'])
                if groups['hms_grp2'] is not None:
                    value += int(groups['hms_grp1']) * 60 * 60 + int(groups['hms_grp2']) * 60
                elif groups['hms_grp1'] is not None:
                    value += int(groups['hms_grp1']) * 60
            else:
                value = float(groups['smu_value'])
                units = groups['smu_units']
                if units == 'ms':
                    value /= 1000.0
                elif units == 'us':
                    value /= 1000000.0
            return value
        else:
            return None

# --------------------------------------------------------------------------------------------------
class DelimitedValueAction(Action):
    """
    An argparse action that splits a list of colon-separated values into a sequence.
    """
    def __init__(self,
        option_strings: Sequence[str],
        dest: str,
        value_type: type=str,
        delimiter: str=':',
        value_choices: Sequence[Any] | None=None,
        nargs=None,
        type=None,
        choices=None,
        **kwargs):

        if nargs is not None:
            msg = 'nargs not allowed'
            raise ValueError(msg)
        if type is not None:
            msg = 'use value_type'
            raise ValueError(msg)
        if choices is not None:
            msg = 'use value_choices'
            raise ValueError(msg)
        self._value_type = value_type
        self._delimiter = delimiter
        self._value_choices = value_choices

        super().__init__(option_strings, dest, type=type, **kwargs)

    # ----------------------------------------------------------------------------------------------
    def __call__(self, parser, ns, values, option_string=None):
        try:
            result = [None if s == '' else self._value_type(s)
                for s in values.split(self._delimiter)]
        except ValueError as ex:
            type_name = self._value_type.__name__
            msg = f"values must be a list of {type_name} values delimited by '{self._delimiter}'"
            raise ArgumentError(self, msg) from ex

        if result is not None and self._value_choices is not None:
            for bad_choice in [choice for choice in result
                if choice is not None and choice not in self._value_choices]:
                    msg = f"invalid choice: '{bad_choice}' (choose from {self._value_choices})"
                    raise ArgumentError(self, msg)
        setattr(ns, self.dest, result)

# --------------------------------------------------------------------------------------------------
class ExtraArgumentParser(ArgumentParser):
    """
    An `argparse.ArgumentParser` subclass that supports a series of arguments which are not parsed.
    """
    def parse_args(
        self,
        args: Sequence[str] | None=None,
        namespace=None,
        marker: str='--',
        dest: str | None=None) -> Namespace:
        """
        Convert argument strings to objects and assign them as attributes of the namespace.  Assign
        arguments after the specified marker to the specified destination.  Return the populated
        namespace.
        """
        # Use default if there is no marker or destination.
        if marker is None or dest is None:
            return super().parse_args(None, namespace)

        # Use default if there is no marker present.
        largs = sys.argv[1:] if args is None else args
        if largs is None or marker not in largs:
            return super().parse_args(None, namespace)

        # Grab all the arguments after the marker arg.
        idx = largs.index(marker)
        unparsed = largs[idx + 1:]

        # Give the base class the rest and assign unparsed args to destination.
        parsed = super().parse_args(largs[:idx], namespace)
        vars(parsed)[dest] = unparsed

        return parsed

# --------------------------------------------------------------------------------------------------
class ConverterArgumentParser(ExtraArgumentParser):
    """
    A `PassthroughArgumentParser` subclass that adds methods for adding shared arguments for
    converter tools in this package.
    """
    def __init__(self, description: str | None):
        """
        Construct an instance with the specified description.
        """
        super().__init__(description=description, fromfile_prefix_chars='@')
        self._has_passthrough_arguments = False
        self._containers : list[Container] = []

    # ----------------------------------------------------------------------------------------------
    def add_basic_arguments(self, group: _ArgumentGroup | None=None) -> None:
        """
        Add basic arguments that apply to all converter tools.
        """
        parent = self if group is None else group
        parent.add_argument('--version', action='version', version='%(prog)s ' + __version__)
        parent.add_argument('-#', '--always-number',
            help='always add a number to the output file name',
            action='store_true', default=False)
        parent.add_argument('--pretend',
            help='display command lines but do not execute',
            action='store_true')
        parent.add_argument('-v', '--verbose',
            help='verbose output',
            action='count', default=0)
        parent.add_argument('source_files',
            help='source files to convert',
            action='store', metavar='source_file', nargs='+')

    # ----------------------------------------------------------------------------------------------
    def add_audio_quality_argument(
            self, format: AudioFormat, group: _ArgumentGroup | None=None) -> None:
        """
        Add an argument quality argument for the specified audio format.
        """
        parent = self if group is None else group
        # Note, I couldn't find a way to pass the options as a list, thus the fixed 'short_flag' and
        # 'long_flag' attributes.
        parent.add_argument(
            format.quality_arg.short_flag,
            format.quality_arg.long_flag,
            help=f'{format.quality_arg.help} (default {format.quality_arg.default})',
            action=DelimitedValueAction,
            dest='audio_quality',
            metavar=format.quality_arg.metavar,
            value_type=format.quality_arg.value_type,
            default=[format.quality_arg.default])

    # ----------------------------------------------------------------------------------------------
    def add_channel_layout_fix_argument(self, group: _ArgumentGroup | None=None) -> None:
        """
        Add a --channel-layout-fix argument.
        """
        parent = self if group is None else group
        parent.add_argument('--channel-layout-fix',
            help='apply a channel layout fix; choices are 4.1, 5.0, 5.1',
            action=DelimitedValueAction, metavar="FIX_STRING",
            value_choices=['0', '4.1', '5.0', '5.1'], default=['0'])

    # ----------------------------------------------------------------------------------------------
    def add_audio_quality_argument_group(self, format: AudioFormat) -> _ArgumentGroup:
        """
        Add a new audio group containing the audio quality argument and, if needed, the channel
        layout fix argument.
        """
        text = (
            'Audio quality is a colon-delimited list of values, where each index matches a source '
            'audio track.  A zero or blank value ignores a source track.  If supported by the '
            'container format, multiple non-empty values may be provided to convert more than one '
            'source track.  By default, and if a single value is provided, only the first audio '
            'track is converted.  For example, 0:x:y or :x:y would ignore track 1, convert track 2 '
            'with  quality x, and convert track 3 with quality y.'
        )
        if format.requires_channel_layout_fix:
            text += (
                '  The channel layout fix works in a similar fashion, and is used to apply a '
                'channel mapping to a layout compatible with the audio codec.  For example, '
                '5.1(side)  is mapped to 5.1(rear).  Specify the channel layout of the source '
                'track.  For  example, use 0:5.1:0 to apply a fix to track 2 containing 5.1(side) '
                'audio.'
            )
        group = self.add_argument_group('audio quality', text)
        self.add_audio_quality_argument(format, group)
        if format.requires_channel_layout_fix:
            self.add_channel_layout_fix_argument(group)

    # ----------------------------------------------------------------------------------------------
    def add_timecode_arguments(self, group: _ArgumentGroup | None=None) -> None:
        """
        Add time code arguments (--start, --end, --duration, --segment).
        """
        parent = self if group is None else group
        parent.add_argument('--start',
            help='starting source position',
            action='store')
        parent.add_argument('--duration',
            help='duration to encode',
            action='store')
        parent.add_argument('--end',
            help='ending source position',
            action='store')
        parent.add_argument('--segment',
            help='segment start and end source position; may be specified multiple times to encode '
                'multiple segments to separate files; enables --always-number when specified more '
                'than once',
            nargs=2, metavar=('START', 'END'), action='append', dest='segments')

    # ----------------------------------------------------------------------------------------------
    def add_timecode_argument_group(self) -> _ArgumentGroup:
        """
        Add a new group containing time code arguments (--start, --end, --duration, --segment) and
        return the group.
        """
        group = self.add_argument_group('source segment arguments',
            'A single segment or multiple segments of a source file may be encoded using the '
            'following arguments.  The first three may be used independently or combined, while '
            'the  last not not be combined with the first three.  The same arguments will be '
            'applied to  all source files.  All argument values are in ffmpeg duration format; see '
            'ffmpeg documentation for more details.')
        self.add_timecode_arguments(group)
        return group

    # ----------------------------------------------------------------------------------------------
    def add_fade_arguments(self, group: _ArgumentGroup | None=None) -> None:
        """
        Add --fade-in and --fade-out arguments.
        """
        parent = self if group is None else group
        parent.add_argument('--fade-in',
            help='apply a fade-in at the start of each output',
            action='store', type=float, metavar='SECONDS')
        parent.add_argument('--fade-out',
            help='apply a fade-out at the end of each output',
            action='store', type=float, metavar='SECONDS')

    # ----------------------------------------------------------------------------------------------
    def add_audio_filter_arguments(
            self, format: AudioFormat, group: _ArgumentGroup | None=None) -> None:
        """
        Add audio filter arguments.
        """
        parent = self if group is None else group
        parent.add_argument('--volume',
            help='amplitude (volume) multiplier, < 1.0 to reduce volume, or > 1.0 to increase '
                'volume; recommended to use replaygain to tag the file post-conversion, instead',
            action='store', type=float, default=1.0)
        parent.add_argument('-af', '--audio-filter',
            help='custom audio filter, passed as -af argument to ffmpeg',
            action='append', dest='audio_filter')

    # ----------------------------------------------------------------------------------------------
    def add_passthrough_argument_group(self) -> _ArgumentGroup:
        """
        Add a new group containing a dummy passthrough argument and return the group.
        """
        # Add a group with additional explanation.
        group = self.add_argument_group('passthrough arguments',
            'Additional arguments can be passed to ffmpeg as-is before the output file name by '
            'adding an ''--'' argument followed by the ffmpeg arguments.  Note, because these '
            'preceed the output file name, they are only useful for output arguments.  Must appear '
            'as the final arguments.')

        # Add a dummy argument that defines a destination and adds help/usage text.  We will pull
        # this and following arguments from those passed to the base class.  Unfortunately, the
        # usage text will show this before any positional arguments.
        group.add_argument('--',
            help='passthrough output arguments',
            dest='passthrough_args', metavar='ARG', nargs='*', default=[])

        self._has_passthrough_arguments = True
        return group

    # ----------------------------------------------------------------------------------------------
    def add_container_argument(
            self, containers: Sequence[Container], group: _ArgumentGroup | None=None ) -> None:
        """
        Add an argument for selection of the container format.
        """
        self._containers = containers
        if len(containers) > 1:
            parent = self if group is None else group
            choices = [c.name for c in containers]
            parent.add_argument('-C', '--container',
                help=f'container format (default {choices[0]})',
                action='store', choices=choices, default=choices[0])

    # ----------------------------------------------------------------------------------------------
    def _check_timecode_arguments(self, args: Namespace) -> None:
        """
        Raise a parser error if the specified args contain an invalid combination of arguments
        added by `add_timecode_arguments()`.  Sets `always_number` if any --segment arguments.
        """
        if ('duration' in args and args.duration is not None and
            'end' in args and args.end is not None):
            self.error('--duration and --end may not be used together')

        if 'segments' in args and args.segments is not None:
            if (('start' in args and args.start is not None) or
                ('duration' in args and args.duration is not None) or
                ('end' in args and args.end is not None)):
                self.error('--segments may not be used with other segment selectors')
            if len(args.segments) > 1:
                args.always_number = True

        if 'fade_out' in args and args.fade_out is not None:
            if not (('duration' in args and args.duration is not None) or
                ('end' in args and args.end is not None) or
                ('segments' in args and args.segments is not None)):
                self.error('--fade-out requires --duration, --end, or --segment')

    # ----------------------------------------------------------------------------------------------
    def _check_source_files_exist(self, source_files: Sequence[str]) -> None:
        """
        Raise a parser error if the specified list contains an element that is not an existing
        file.
        """
        for source_file in source_files:
            if not os.path.exists(source_file) or not os.path.isfile(source_file):
                self.error('invalid source file: ' + source_file)

    # ----------------------------------------------------------------------------------------------
    def parse_args(self, args: Sequence[str] | None=None, namespace=None) -> Namespace:
        """
        Convert argument strings to objects and assign them as attributes of the namespace.  Return
        the populated namespace.
        """
        dest = 'passthrough_args' if self._has_passthrough_arguments else None
        parsed = super().parse_args(args, namespace, dest=dest)
        self._check_timecode_arguments(parsed)

        # Add/replace `container` with the original Container instance from _containers.
        if (len(self._containers) == 1):
            parsed.container = self._containers[0]
        else:
            parsed.container = next(f for f in self._containers if f.name == parsed.container)

        # Create an `all_segments` containing Segment instances representing all timecode args.
        if parsed.segments is not None:
            parsed.all_segments = [Segment(s[0], s[1], None) for s in parsed.segments]
        else:
            parsed.all_segments = [Segment(parsed.start, parsed.end, parsed.duration)]

        if 'source_files' in parsed and parsed.source_files is not None:
            self._check_source_files_exist(parsed.source_files)

        return parsed

# --------------------------------------------------------------------------------------------------
class AudioConverterArgumentParser(ConverterArgumentParser):
    """
    A `ConverterArgumentParser` subclass for audio converter tools.
    """
    def __init__(self, audio_format: AudioFormat):
        """
        Construct an argument parser pre-populated with arguments for an audio converter tool that
        outputs the specified format.
        """
        self.audio_format = audio_format
        desc = f'Converts audio/video files to audio-only {audio_format.name} using ffmpeg'
        super().__init__(desc)

        # Options group
        self.add_basic_arguments()
        # Note: This may not add an argument if there is only one container chice, but we need to
        # call it anyway.
        self.add_container_argument(audio_format.containers)

        # Audio quality group
        self.add_audio_quality_argument_group(audio_format)

        # Timecode group
        self.add_timecode_argument_group()

        # Filter group
        group = self.add_argument_group('filter arguments')
        self.add_fade_arguments(group)
        self.add_audio_filter_arguments(audio_format, group)

        # Passthrough
        self.add_passthrough_argument_group()

    # ----------------------------------------------------------------------------------------------
    def parse_args(self, args: Sequence[str] | None=None, namespace=None) -> Namespace:
        """
        Convert argument strings to objects and assign them as attributes of the namespace.  Return
        the populated namespace.
        """
        parsed = super().parse_args(args, namespace)

        qual_name = self.audio_format.quality_arg.metavar
        if parsed.container.supports_multiple_tracks:
            if len([q for q in parsed.audio_quality if q is not None and q > 0]) < 1:
                msg = f'at least one positive audio {qual_name} must be specified'
                self.error(msg)
        elif len([q for q in parsed.audio_quality if q is not None and q > 0]) != 1:
            msg = f'exactly one non-zero {qual_name} must be specified for the container'
            self.error(msg)

        return parsed

# --------------------------------------------------------------------------------------------------
class VideoConverterArgumentParser(ConverterArgumentParser):
    def __init__(self, video_format: VideoFormat):
        """
        Construct an argument parser pre-populated with arguments for a video converter tool that
        outputs the specified format.
        """
        self.video_format = video_format
        s = ' or '.join([f.name for f in video_format.containers])
        desc = (
            f'Converts video files to {video_format.name} + '
            f'{video_format.audio_format.name} in a {s} container using a '
            f'{video_format.passes}-pass ffmpeg encode.'
        )
        super().__init__(desc)

        # Options group
        self.add_basic_arguments()
        self.add_mapped_codec_args()
        # Note: This may not add an argument if there is only one container chice, but we need to
        # call it anyway.
        self.add_container_argument(video_format.containers)

        # Audio quality group
        self.add_audio_quality_argument_group(video_format.audio_format)

        # Add arguments only needed if more than one pass.  Note: 'pass' is a keyword, so
        # 'pass_num' is used internally.
        if video_format.passes > 1:
            self.add_multi_pass_arguments(video_format.passes)

        # Timecode group.
        self.add_timecode_argument_group()

        # Video filter group.
        group = self.add_argument_group('video filter arguments',
            'The deinterlate filter is applied first; standard crop filters (e.g., "-s crop43") are '
            'applied before custom crop values (e.g., "-x 10 10"); crop filters are applied before '
            'scale filters (e.g., "-s scale23");  and all standard filters are applied before custom '
            '-vf filters.')
        self.add_video_filter_arguments(group)

        # Audio filter group.
        group = self.add_argument_group('audio filter arguments')
        self.add_audio_filter_arguments(video_format.audio_format, group)

        # Passthrough
        self.add_passthrough_argument_group()

    # ----------------------------------------------------------------------------------------------
    def add_video_quality_argument(self, group: _ArgumentGroup | None=None) -> None:
        """
        Add the --quality argument.
        """
        parent = self if group is None else group
        s = f'{self.video_format.video_quality_help} (default {self.video_format.default_quality})'
        parent.add_argument('-q', '--quality',
            help=s, action='store', type=int, default=self.video_format.default_quality)

    # ----------------------------------------------------------------------------------------------
    def add_mapped_codec_arg(self, mapped_arg: MappedCodecArg, group: _ArgumentGroup | None=None) -> None:
        """
        Add a mapped codec arg.
        """
        parent = self if group is None else group
        parent.add_argument(
            mapped_arg.short_flag,
            mapped_arg.long_flag,
            help=f'{mapped_arg.help} (default {mapped_arg.default})',
            action='store',
            type=mapped_arg.value_type,
            default=mapped_arg.default,
            dest=mapped_arg.dest
        )

    # ----------------------------------------------------------------------------------------------
    def add_mapped_codec_args(self, group: _ArgumentGroup | None=None) -> None:
        """
        Add mapped codec args from the video format definition.
        """
        parent = self if group is None else group
        for mapped_arg in self.video_format.mapped_codec_args:
            self.add_mapped_codec_arg(mapped_arg, parent)

    # ----------------------------------------------------------------------------------------------
    def add_video_filter_arguments(self, group: _ArgumentGroup | None=None):
        """
        Add video filter arguments.
        """
        parent = self if group is None else group
        parent.add_argument('-s', '--standard-filter',
            help='standard video/audio filter; '
                '[crop43] crops horizontally to 4:3 aspect ratio; '
                '[scale23] scales down by 2/3 (e.g., 1080p to 720p); '
                '[gray] converts to grayscale',
            action='append', choices=['crop43', 'scale23', 'gray'])
        parent.add_argument('-d', '--deinterlace',
            help='deinterlate filter; '
                '[frame] output a frame from each pair of input fields; '
                '[field] output an interpolated frame from each input field; '
                '[ivtc] inverse telecine; '
                '[ivtc+] inverse telecine with fallback deinterlace; '
                '[selframe] selectively deinterlace frames ',
            action='store', choices=['frame', 'field', 'ivtc', 'ivtc+', 'selframe'])
        parent.add_argument('--parity',
            help='set a specific parity for the deinterlace filter; '
                '[tff] top field first; '
                '[bff] bottom field first',
            action='store', choices=['tff', 'bff'])

        self.add_fade_arguments(parent)

        parent.add_argument('-x', '--crop-width',
            help='left and right crop values',
            nargs=2, type=int, metavar=('LEFT', 'RIGHT'))
        parent.add_argument('-y', '--crop-height',
            help='top and bottom crop values',
            nargs=2, type=int, metavar=('TOP', 'BOTTOM'))
        parent.add_argument('-g', '--gamma',
            help='gramma correction (default 1.0; no correction)',
            action='store', type=float, default=1.0)
        parent.add_argument('-vf', '--video-filter',
            help='custom video filter, similar to -vf ffmpeg argument',
            action='append')

    # ----------------------------------------------------------------------------------------------
    def add_multi_pass_arguments(self, passes: list[int], group: _ArgumentGroup | None=None):
        """
        Add arguments that relate to a multi-pass encode.
        """
        parent = self if group is None else group
        parent.add_argument('--pass',
            help='run only a given pass',
            action='store', choices=range(1, passes + 1), dest='pass_num', type=int)
        parent.add_argument('--delete-log',
            help='delete pass 1 log (otherwise keep with timestamp)',
            action='store_true')



# formats.py - Audio/video format definitions.
# Copyright (C) 2025 David Gasaway
# https://github.com/dgasaway/towebm

# This program is free software; you can redistribute it and/or modify it under the terms of the GNU
# General Public License as published by the Free Software Foundation; either version 2 of the
# License, or (at your option) any later version.
#
# This program is distributed in the hope that it will be useful, but WITHOUT ANY WARRANTY; without
# even the implied warranty of MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE. See the GNU
# General Public License for more details.
#
# You should have received a copy of the GNU General Public License along with this program; if not,
# see <http://www.gnu.org/licenses>.
from __future__ import annotations

from dataclasses import dataclass
from typing import Final

# --------------------------------------------------------------------------------------------------
@dataclass
class MappedCodecArg:
    """
    Represents an tool argument that maps directly to an ffmpeg codec arg.
    """
    short_flag: str | None
    """
    The short option string.
    """
    long_flag: str | None
    """
    The long option string.
    """
    ffmpeg_arg: str
    """
    The argument to pass to ffmpeg.
    """
    dest: str
    """
    The name of the attribute to be added to the namespace.
    """
    help: str = None
    """
    A brief description of what the argument does.
    """
    value_type: type = str
    """
    The type to which the command-line argument should be converted.
    """
    default: any = None
    """
    The default value for the argument.
    """
    metavar: str | None = None
    """
    A name for the argument in usage messages.
    """

# --------------------------------------------------------------------------------------------------
@dataclass
class AudioQualityArg(MappedCodecArg):
    """
    Represents an audio quality argument to add to a Converter CLI.
    """
    ffmpeg_value_suffix: str | None = None
    """
    The suffix to add to the quality value for the ffmpeg arg, if needed.
    """

# --------------------------------------------------------------------------------------------------
@dataclass
class Container:
    """
    Represents the attributes of a container format.
    """
    name: str
    """
    The name of the container format.
    """
    ffmpeg_format: str
    """
    The ffmpeg format name of the container format.
    """
    extension: str
    """
    The file extension to add for the container format.
    """
    supports_multiple_tracks: bool
    """
    True if the format supports multiple audio tracks, otherwise False.
    """

# --------------------------------------------------------------------------------------------------
class Containers:
    """
    Contains static `Container` instances of the defined container formats.
    """
    OGG: Final[Container] = Container("Ogg", "ogg", ".ogg", True)
    FLAC: Final[Container] = Container("FLAC", "flac", ".flac", False)
    MKV: Final[Container] = Container("Matroska", "matroska", ".mkv", True)
    WEBM: Final[Container] = Container("WebM", "webm", ".webm", True)
    MP4: Final[Container] = Container("MP4", "mp4", ".mp4", True)
    MP3: Final[Container] = Container("MP3", "mp3", ".mp3", False)

# --------------------------------------------------------------------------------------------------
@dataclass
class Format:
    """
    Represents the attributes of an audio or video format.
    """
    name: str
    """
    The name of the audio format.
    """
    ffmpeg_codec: str
    """
    The name of the corresponding ffmpeg codec.
    """
    containers: list[Container]
    """
    The list of supported container formats when the output is an audio file.
    """

# --------------------------------------------------------------------------------------------------
@dataclass
class AudioFormat(Format):
    """
    Represents the attributes of an audio format.
    """
    quality_arg: AudioQualityArg
    """
    The audio quality argument definition.
    """
    requires_channel_layout_fix: bool
    """
    True if the format requires mapping 5.1(side) to 5.1(rear) or similar for 5.0/4.1.
    """
    codec_args: list[str]
    """
    The ffmpeg codec arguments to add.  Arguments in index 0 are added for all passes, arguments in
    index 1 are added to pass 1, and so on.
    """

# --------------------------------------------------------------------------------------------------
class AudioFormats:
    """
    Contains static `AudioFormat` instances of the defined audio formts.
    """
    # vorbis output is not picky about channel layout.
    VORBIS: Final[AudioFormat] = AudioFormat(
        'Vorbis', 'libvorbis', [Containers.OGG],
        AudioQualityArg('-q', '--quality', '-q',
            dest='audio_quality',
            help='audio quality',
            value_type=float,
            default=6.0
        ),
        False,
        [ ]
    )
    # ffmpeg will create a multi-track opus file.  VLC will not play it; mplayer will.
    OPUS: Final[AudioFormat] = AudioFormat(
        'Opus', 'libopus', [Containers.OGG],
        AudioQualityArg('-b', '--bitrate', '-b',
            dest='audio_quality',
            help='audio bitrate in kbps',
            value_type=int,
            default=160,
            ffmpeg_value_suffix='k'
        ),
        True,
        [ ]
    )
    FLAC: Final[AudioFormat] = AudioFormat(
        'FLAC', 'flac', [Containers.FLAC, Containers.OGG],
        AudioQualityArg('-c', '--compression-level', '-compression_level',
            dest='audio_quality',
            help='audio compression level',
            value_type=int,
            default=8,
            metavar='COMP_LEVEL',
        ),
        False,
        [ ]
    )
    MP3: Final[AudioFormat] = AudioFormat(
        'MP3', 'libmp3lame', [Containers.MP3],
        AudioQualityArg('-q', '--quality', '-q',
            dest='audio_quality',
            help='audio quality in LAME VBR mode',
            value_type=float,
            default=4.0
        ),
        False,
        ['-compression_level', '0']
    )

# --------------------------------------------------------------------------------------------------
@dataclass
class VideoFormat(Format):
    """
    Represents the attributes of a video format.
    """
    passes: int
    """
    The number of encode passes for this format and codec.
    """
    audio_format: AudioFormat
    """
    The audio format to use for this video format.
    """
    mapped_codec_args: list[MappedCodecArg]
    """
    The list of mapped codec args.
    """
    codec_args: list[list[str]]
    """
    The ffmpeg codec arguments to add.  Arguments in index 0 are added for all passes, arguments in
    index 1 are added to pass 1, and so on.
    """

# --------------------------------------------------------------------------------------------------
class VideoFormats:
    """
    Contains static `VideoFormat` instances of the defined video formats.
    """
    WEBM: Final[VideoFormat] = VideoFormat(
        'VP9',
        'libvpx-vp9',
        [Containers.WEBM, Containers.MKV, Containers.MP4],
        2,
        AudioFormats.OPUS,
        [
            MappedCodecArg('-q', '--quality', '-crf',
                dest='video_quality',
                help='video quality, lower is better',
                value_type=int,
                default=30
            )
        ],
        [
            # All pass args.
            [
                '-b:v', '0',
                '-tile-columns', '2',
                '-row-mt', '1',
                '-auto-alt-ref', '1',
                '-lag-in-frames', '25',
                '-threads', '8',
                '-pix_fmt', 'yuv420p'
            ],
            # Pass one args.
            ['-cpu-used', '4'],
            # Pass two args.
            ['-cpu-used', '2']
        ]
    )

    AV1_SVT: Final[VideoFormat] = VideoFormat(
        'AV1',
        'libsvtav1',
        [Containers.MKV, Containers.MP4],
        1,
        AudioFormats.OPUS,
        [
            MappedCodecArg('-q', '--quality', '-crf',
                dest='video_quality',
                help='video quality, lower is better',
                value_type=int,
                default=30
            ),
            MappedCodecArg('-p', '--preset', '-preset',
                dest='preset',
                help='AV1-SVT preset; lower improves quality, higher improves speed',
                value_type=int,
                default=3
            )
        ],
        [ ]
    )

    AV1_AOM: Final[VideoFormat] = VideoFormat(
        'AV1',
        'libaom-av1',
        [Containers.MKV, Containers.MP4],
        2,
        AudioFormats.OPUS,
        [
            MappedCodecArg('-q', '--quality', '-crf',
                dest='video_quality',
                help='video quality, lower is better',
                value_type=int,
                default=30
            )
        ],
        [
            # All pass args.
            [
                '-b:v', '0',
                '-tile-columns', '2',
                '-row-mt', '1',
                '-auto-alt-ref', '1',
                '-lag-in-frames', '25'
            ],
            # Pass one args.
            ['-cpu-used', '4'],
            # Pass two args.
            ['-cpu-used', '2']
        ],
    )



# ffcat - Concatenates media files using ffmpeg.
# Copyright (C) 2025 David Gasaway
# https://github.com/dgasaway/towebm

# This program is free software; you can redistribute it and/or modify it under the terms of the GNU
# General Public License as published by the Free Software Foundation; either version 2 of the
# License, or (at your option) any later version.
#
# This program is distributed in the hope that it will be useful, but WITHOUT ANY WARRANTY; without
# even the implied warranty of MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE. See the GNU
# General Public License for more details.
#
# You should have received a copy of the GNU General Public License along with this program; if not,
# see <http://www.gnu.org/licenses>.
from __future__ import annotations

import os
import subprocess
import sys
from argparse import ArgumentParser, Namespace
from tempfile import NamedTemporaryFile

from towebm._version import __version__

# --------------------------------------------------------------------------------------------------
def main() -> int:
    """
    Executes the operations indicated by the command line arguments.
    """
    rc = 0
    args = parse_args()
    file_list = NamedTemporaryFile(mode='wt', dir=os.getcwd(), delete=False)
    try:
        for source_file in args.source_files:
            # File names will be in single quotes with existing quotes padded.
            name = source_file.replace("'", r"'\''")
            file_list.write(f"file '{name}'\n")
        file_list.close()
        ffmpeg_args = ['ffmpeg', '-f', 'concat', '-safe', '0', '-i', file_list.name, '-c', 'copy',
                       args.output_file]
        if args.verbose >= 1:
            print(ffmpeg_args)
        subprocess.check_call(ffmpeg_args)
    except subprocess.CalledProcessError as ex:
        rc = ex.returncode
    finally:
        os.remove(file_list.name)

    return rc

# --------------------------------------------------------------------------------------------------
def parse_args() -> Namespace:
    """
    Parses and returns the command line arguments.
    """
    parser = ArgumentParser(
        description='Concatenates media files using the ffmpeg concat demuxer.',
        fromfile_prefix_chars='@')
    parser.add_argument('--version', action='version', version='%(prog)s ' + __version__)
    parser.add_argument('-v', '--verbose',
        help='verbose output',
        action='store_true')
    parser.add_argument('source_files',
        help='sources files to concatenate',
        action='store', metavar='source_file', nargs='+')
    parser.add_argument('output_file',
        help='output file name', action='store')
    args = parser.parse_args()

    if args.verbose >= 1:
        print(args)

    return args

# --------------------------------------------------------------------------------------------------
if __name__ == "__main__":
     sys.exit(main())



# tovorbis.py - Converts audio/video files to audio-only vorbis using ffmpeg.
# Copyright (C) 2025 David Gasaway
# https://github.com/dgasaway/towebm

# This program is free software; you can redistribute it and/or modify it under the terms of the GNU
# General Public License as published by the Free Software Foundation; either version 2 of the
# License, or (at your option) any later version.
#
# This program is distributed in the hope that it will be useful, but WITHOUT ANY WARRANTY; without
# even the implied warranty of MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE. See the GNU
# General Public License for more details.
#
# You should have received a copy of the GNU General Public License along with this program; if not,
# see <http://www.gnu.org/licenses>.
import sys

from towebm.converters import AudioConverter
from towebm.formats import AudioFormats

# --------------------------------------------------------------------------------------------------
def main() -> int:
    return AudioConverter(AudioFormats.VORBIS).main()

# --------------------------------------------------------------------------------------------------
if __name__ == "__main__":
    sys.exit(main())



# converters.py - Shared routines.
# Copyright (C) 2025 David Gasaway
# https://github.com/dgasaway/towebm

# This program is free software; you can redistribute it and/or modify it under the terms of the GNU
# General Public License as published by the Free Software Foundation; either version 2 of the
# License, or (at your option) any later version.
#
# This program is distributed in the hope that it will be useful, but WITHOUT ANY WARRANTY; without
# even the implied warranty of MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE. See the GNU
# General Public License for more details.
#
# You should have received a copy of the GNU General Public License along with this program; if not,
# see <http://www.gnu.org/licenses>.
from __future__ import annotations

from abc import ABC, abstractmethod
from datetime import datetime
import os
import subprocess
import sys

from typing import TYPE_CHECKING
if TYPE_CHECKING:
    from argparse import Namespace
    from towebm.formats import AudioFormat, VideoFormat
    from towebm.argparsers import Segment

from towebm.argparsers import AudioConverterArgumentParser, VideoConverterArgumentParser

# --------------------------------------------------------------------------------------------------
class Converter(ABC):
    """
    A base class that converts source audio/video source files to audio or video output files based
    on a list of argument strings provided to main().
    """
    def __init__(self, audio_format: AudioFormat):
        self.args: Namespace = None
        self.audio_format = audio_format

    # ----------------------------------------------------------------------------------------------
    @abstractmethod
    def parse_args(self, args: list[str] | None=None) -> Namespace:
        """
        Convert the specified argument strings, or sys.argv if None, to a namespace.  Return the
        populated namespace.
        """

    # ----------------------------------------------------------------------------------------------
    def get_safe_filename(self, filename: str) -> str:
        """
        Return the source file name if no file exists with the given name and the `always_number`
        arg is false; return the source file name with an understore and two-digit sequence number
        appended to make the file name unique if the source file name is not unique or the
        `always_number` arg is true; if no such file name is unique, return the source file name.
        """
        if not self.args.always_number and not os.path.exists(filename):
            return filename
        else:
            (base, ext) = os.path.splitext(filename)
            for i in range(100):
                s = f'{base}_{i:02}{ext}'
                if not os.path.exists(s):
                    return s
            return filename

    # ----------------------------------------------------------------------------------------------
    def get_fade_filters(self, segment: Segment, filter_name: str) -> list[str]:
        """
        Return a list of fade-in and/or fade-filters for the specified arguments based on fade
        argments, or an empty list if none were specified in the arguments.
        """
        # The fade filters take a start time relative to the start of the output, rather than the
        # start of the source.  So, fade in will start at 0 secs.  Fade out needs to get the output
        # duration and subtract the fade out duration.
        filters = []
        if self.args.fade_in is not None:
            filters.append(f'{filter_name}=t=in:st=0:d={self.args.fade_in}')
        if self.args.fade_out is not None and segment.duration is not None:
            start = segment.duration - self.args.fade_out
            filters.append(f'{filter_name}=t=out:st={start}:d={self.args.fade_out}')
        return filters

    # ----------------------------------------------------------------------------------------------
    def get_audio_filters(self, segment: Segment) -> list[str]:
        """
        Return a list of audio filters, one element per standard filter or audio filter arg.
        """
        filters: list[str] = []

        # Want to apply standard filters is a certain order, so do not loop.
        if self.args.volume != 1.0:
            filters.append(f'volume={self.args.volume}')
        filters += self.get_fade_filters(segment, 'afade')
        if self.args.audio_filter is not None:
            filters += self.args.audio_filter

        return filters

    # ----------------------------------------------------------------------------------------------
    def get_audio_filter_args(self, segment: Segment) -> list[str]:
        """
        Return a list of ffmpeg arguments that apply all of the selected audio filters specified
        in the args, or an empty list if none apply.
        """
        filters = self.get_audio_filters(segment)
        per_track_filters = []

        # We need to specify the input index for each that audio stream that will be output.  So, we
        # iterate the list with index, rather than use list comprehension.
        for i, quality in enumerate(self.args.audio_quality):
            if quality is not None and quality > 0:
                # channel_layout_fix is going to use the same index, but it may have fewer values
                # specified than audio_quality.
                map_fix = (
                    'channel_layout_fix' in self.args and
                    i < len(self.args.channel_layout_fix) and
                    self.args.channel_layout_fix[i] is not None and
                    self.args.channel_layout_fix[i] != '0')
                if map_fix:
                    layout = self.args.channel_layout_fix[i]
                    if layout == '5.1':
                        flts = ['channelmap=channel_layout=5.1']
                    elif layout == '5.0':
                        flts = ['pan=5.1|FR=FR|FL=FL|FC=FC|BL=SL|BR=SR']
                    elif layout == '4.1':
                        flts = ['pan=5.1|FR=FR|FL=FL|FC=FC|BL=BC|BR=BC|LFE=LFE']
                    else:
                        msg = f'unexpected channel layout fix value {layout}'
                        raise ValueError(msg)
                    flts += filters
                elif len(filters) == 0:
                    flts = ['acopy']
                else:
                    flts = filters
                per_track_filters.append(f'[0:a:{i}]' + ','.join(flts))

        if len(per_track_filters) == 0:
            return []
        else:
            return ['-filter_complex', ';'.join(per_track_filters)]

    # ----------------------------------------------------------------------------------------------
    def get_segment_arguments(self, segment: Segment) -> list[str]:
        """
        Return a list of ffmepg arguments to select the specified source segment, or an empty list
        if no segment was specified.
        """
        result = []
        # For the args, it's best to use the original ffmpeg strings.
        if segment.start_str is not None:
            result += ['-accurate_seek', '-ss', segment.start_str]
        if segment.end_str is not None:
            result += ['-to', segment.end_str]
        if segment.duration_str is not None:
            result += ['-t', segment.duration_str]
        return result

    # ----------------------------------------------------------------------------------------------
    def get_base_segment_args(self, segment: Segment, file_name: str) -> list[str]:
        """
        Return a list of ffmpeg arguments for transcoding the specified segment of the specified
        file.
        """
        result = ['ffmpeg', '-nostdin', '-hide_banner']
        result += self.get_segment_arguments(segment)
        result += ['-i', file_name]
        result += ['-f', self.args.container.ffmpeg_format]
        return result

    # ----------------------------------------------------------------------------------------------
    def get_audio_quality_arg(self, quality: float, stream_index: int | None=None) -> list[str]:
        """
        Return a list of ffmpeg arguments for a specified audio quality and optional output stream
        index.
        """
        quality_arg = self.audio_format.quality_arg
        arg = f'{quality_arg.ffmpeg_arg}:a'
        if stream_index is not None:
            arg += f':{stream_index}'

        value = str(quality)
        if quality_arg.ffmpeg_value_suffix is not None:
            value += quality_arg.ffmpeg_value_suffix

        return [arg, value]

    # ----------------------------------------------------------------------------------------------
    def get_audio_quality_args(self) -> list[str]:
        """
        Return a list of one or more sets of ffmpeg audio quality arguments based on the audio
        quality arguments in the args.
        """
        # We only output a quality for non-zero values, and since the stream index is the output
        # index, we can use the index of a filtered list.
        result = []
        qualities = [q for q in self.args.audio_quality if q is not None and q > 0]
        for i, quality in enumerate(qualities):
            result += self.get_audio_quality_arg(quality, i)
        return result

    # ----------------------------------------------------------------------------------------------
    def get_audio_metadata_map_arg(self,
        output_index: int=0, input_index: int | None=None) -> list[str]:
        """
        Return a list two ffmpeg arguments for copying audio stream metadata from a specified
        source index to a specified output stream index.
        """
        arg = f'-map_metadata:s:a:{output_index}'
        if input_index is None:
            return [arg, '0:s:a']
        else:
            return [arg, f'0:s:a:{input_index}']

    # --------------------------------------------------------------------------------------------------
    def get_audio_format_args(self) -> list[str]:
        """
        Return a list of the ffmpeg audio codec arguments from the audio format specification.
        """
        codec_args = self.audio_format.codec_args
        if codec_args is not None and len(codec_args) > 0:
            return codec_args
        else:
            return []

    # ----------------------------------------------------------------------------------------------
    def is_iterable(self, obj) -> bool:
        """
        Determine whether the specified object is iterable.
        """
        try:
            iter(obj)
        except TypeError:
            return False
        else:
            return True

    # ----------------------------------------------------------------------------------------------
    def get_audio_metadata_map_args(self) -> list[str]:
        """
        Return a list of ffmpeg arguments to copy audio metadata from the input streams to the
        matching output streams.
        """
        result: list[str] = []
        if self.is_iterable(self.args.audio_quality):
            # We need both the input and output index to create the map.
            output_index = 0
            for input_index, quality in enumerate(self.args.audio_quality):
                if quality is not None and quality > 0:
                    result += self.get_audio_metadata_map_arg(output_index, input_index)
                    output_index += 1
        elif self.args.audio_quality > 0:
            result = self.get_audio_metadata_map_arg()
        return result

    # ----------------------------------------------------------------------------------------------
    @abstractmethod
    def get_segment_commands(self, segment: Segment, file_name: str) -> list[list[str]]:
        """
        Return the list of commands to transcode the specified segment of the specified input file.
        """

    # ----------------------------------------------------------------------------------------------
    def process_segment(self, segment: Segment, file_name: str) -> None:
        """
        Execute the requested action for a single output file.
        """
        commands = self.get_segment_commands(segment, file_name)
        if self.args.pretend or self.args.verbose >= 1:
            for command in commands:
                #print(' '.join(command))
                print(command)
                print()
        if not self.args.pretend:
            for command in commands:
                subprocess.check_call(command)

    # ----------------------------------------------------------------------------------------------
    def process_file(self, file_name: str) -> None:
        """
        Executes the requested action for a single input file.
        """
        for segment in self.args.all_segments:
            self.process_segment(segment, file_name)

    # ----------------------------------------------------------------------------------------------
    def setup(self, args: list[str] | None=None) -> int:
        """
        Convert the specified argument strings, or sys.argv if None, to a namespace and assign then
        to `self.args`.  If `self.args` is set, issues a warning.
        """
        if args is not None:
            print('WARNING: overriding args', file=sys.srderr)
        self.args = self.parse_args(args)
        if self.args.verbose >= 2:
            print (self.args)

    # ----------------------------------------------------------------------------------------------
    def main(self, args: list[str] | None=None) -> int:
        """
        Execute the operations indicated by specified argument strings, or sys.argv if None.
        """
        # We'll treat each input file as it's own job, and continue to the next if there is a
        # problem.
        self.setup(args)
        rc = 0
        for source_file in self.args.source_files:
            try:
                self.process_file(source_file)
            except subprocess.CalledProcessError as ex:
                if rc == 0 or ex.returncode > rc:
                    rc = ex.returncode
                print('Execution error, proceeding to next source file.', file=sys.stderr)

        return rc

# --------------------------------------------------------------------------------------------------
class AudioConverter(Converter):
    """
    An audio converter.
    """
    # ----------------------------------------------------------------------------------------------
    def __init__(self, audio_format: AudioFormat):
        super().__init__(audio_format)

    # ----------------------------------------------------------------------------------------------
    def parse_args(self, args: list[str] | None=None) -> Namespace:
        """
        Convert the specified argument strings, or sys.argv if None, to a namespace.  Return the
        populated namespace.
        """
        return AudioConverterArgumentParser(self.audio_format).parse_args(args)

    # ----------------------------------------------------------------------------------------------
    def get_segment_command(self, segment: Segment, file_name: str) -> list[str]:
        """
        Return the arguments to run ffmpeg for a single output file.
        """
        title = os.path.splitext(os.path.basename(file_name))[0]

        result = self.get_base_segment_args(segment, file_name)
        result += [
            '-vn',
            '-c:a', self.audio_format.ffmpeg_codec
        ]
        result += self.get_audio_filter_args(segment)
        result += self.get_audio_quality_args()
        result += self.get_audio_format_args()
        result += self.get_audio_metadata_map_args()
        result += self.args.passthrough_args

        safe_file_name = self.get_safe_filename(f'{title}{self.args.container.extension}')
        result.append(safe_file_name)

        return result

    # ----------------------------------------------------------------------------------------------
    def get_segment_commands(self, segment: Segment, file_name: str) -> list[list[str]]:
        """
        Return the list of commands to transcode the specified segment of the specified input file.
        """
        return [self.get_segment_command(segment, file_name)]

# --------------------------------------------------------------------------------------------------
class VideoConverter(Converter):
    """
    """
    # ----------------------------------------------------------------------------------------------
    def __init__(self, video_format: VideoFormat):
        super().__init__(video_format.audio_format)
        self.video_format = video_format

    # ----------------------------------------------------------------------------------------------
    def parse_args(self, args: list[str] | None=None) -> Namespace:
        """
        Convert the specified argument strings, or sys.argv if None, to a namespace.  Return the
        populated namespace.
        """
        return VideoConverterArgumentParser(self.video_format).parse_args(args)

    # ----------------------------------------------------------------------------------------------
    def get_video_filter_args(self, segment: Segment) -> list[str]:
        """
        Return a list of ffmpeg arguments that apply all of the selected video filters requested in
        the specified script arguments, or an empty list if none apply.
        """
        filters: list[str] = []

        # Deinterlace first.
        parity = ''
        if self.args.parity is not None:
            parity = ':' + self.args.parity
        if self.args.deinterlace == 'frame':
            filters.append('bwdif=send_frame' + parity)
        elif self.args.deinterlace == 'field':
            filters.append('bwdif=send_field' + parity)
        elif self.args.deinterlace == 'ivtc':
            filters += ['fieldmatch', 'decimate']
        elif self.args.deinterlace == 'ivtc+':
            filters += ['fieldmatch', 'bwdif=send_frame', 'decimate']
        elif self.args.deinterlace == 'selframe':
            filters += ['fieldmatch', 'bwdif=0:-1:1']

        # Want to apply standard filters is a certain order, so do not loop.
        if self.args.standard_filter is not None:
            if 'gray' in self.args.standard_filter:
                filters.append('format=gray')
            if 'crop43' in self.args.standard_filter:
                filters.append('crop=w=(in_h*4/3)')

        if self.args.gamma != 1.0:
            filters.append(f'eq=gamma={self.args.gamma}')

        if self.args.crop_width is not None or self.args.crop_height is not None:
            if self.args.crop_width is not None and self.args.crop_height is not None:
                crop = 'crop=x={x[0]}:w=in_w-{x[0]}-{x[1]}:y={y[0]}:h=in_h-{y[0]}-{y[1]}'
            elif self.args.crop_width is not None:
                crop = 'crop=x={x[0]}:w=in_w-{x[0]}-{x[1]}'
            else:
                crop = 'crop=y={y[0]}:h=in_h-{y[0]}-{y[1]}'
            filters.append(crop.format(x=self.args.crop_width, y=self.args.crop_height))

        if self.args.standard_filter is not None:
            if 'scale23' in self.args.standard_filter:
                filters.append('scale=h=in_h*2/3:w=-1')

        filters += self.get_fade_filters(segment, 'fade')

        if self.args.video_filter is not None:
            filters += self.args.video_filter

        if len(filters) == 0:
            filters = ['copy']

        return ['-filter_complex', '[0:v]' + ','.join(filters)]

    # --------------------------------------------------------------------------------------------------
    def get_audio_codec_args(self, segment: Segment) -> list[str]:
        """
        Return a list of ffmpeg audio codec and audio filter arguments for the specified segment.
        """
        if len([q for q in self.args.audio_quality if q is not None and q > 0]) > 0:
            result = ['-c:a', self.video_format.audio_format.ffmpeg_codec]
            result += self.get_audio_filter_args(segment)
            result += self.get_audio_format_args()
            result += self.get_audio_quality_args()
        else:
            result = ['-an']

        return result

    # --------------------------------------------------------------------------------------------------
    def get_video_format_args(self, pass_num: int) -> list[str]:
        """
        Return a list of the ffmpeg video codec arguments from the video format specification for the
        specified pass.
        """
        codec_args = self.video_format.codec_args
        result = []
        if codec_args is not None and len(codec_args) > 0:
            if codec_args[0] is not None:
                result.extend(codec_args[0])
            if len(codec_args) > pass_num and codec_args[pass_num] is not None:
                result.extend(codec_args[pass_num])
        for mapped_arg in self.video_format.mapped_codec_args:
            result.append(mapped_arg.ffmpeg_arg)
            result.append(str(getattr(self.args, mapped_arg.dest)))

        return result

    # --------------------------------------------------------------------------------------------------
    def get_video_codec_args(self, segment: Segment, pass_num: int) -> list[str]:
        """
        Return a list of ffmpeg video codec and video filter arguments for the specified segment and
        pass number.
        """
        result = [
            '-c:v', self.video_format.ffmpeg_codec,
            #self.video_format.video_quality_arg, str(self.args.quality)
        ]
        result += self.get_video_format_args(pass_num)
        result += self.get_video_filter_args(segment)

        return result

    # --------------------------------------------------------------------------------------------------
    def get_one_pass_command(self, segment: Segment, file_name: str) -> list[str]:
        """
        Return a complete ffmpeg command and arguments to perform a one-pass transcode of the specified
        segment and input file.
        """
        title = os.path.splitext(os.path.basename(file_name))[0]

        result = self.get_base_segment_args(segment, file_name)
        result += self.get_video_codec_args(segment, 1)
        result += self.get_audio_codec_args(segment)
        result += ['-metadata', f'title={title}']
        result += self.get_audio_metadata_map_args()
        result += self.args.passthrough_args

        safe_file_name = self.get_safe_filename(f'{title}{self.args.container.extension}')
        result.append(safe_file_name)

        return result

    # --------------------------------------------------------------------------------------------------
    def get_not_last_pass_command(self, segment: Segment, file_name: str, pass_num: int) -> list[str]:
        """
        Return a complete ffmpeg command and arguments to perform the specified video-only no-output
        pass for the specified segment and input file.
        """
        title = os.path.splitext(os.path.basename(file_name))[0]

        result = self.get_base_segment_args(segment, file_name)
        result += self.get_video_codec_args(segment, pass_num)
        result += [
            # No audio.
            '-an',
            # Always overwrite an existing log file.
            '-y',
            '-pass', str(pass_num),
            '-passlogfile', title,
        ]
        result += self.args.passthrough_args
        result.append('/dev/null')

        return result

    # --------------------------------------------------------------------------------------------------
    def get_last_pass_command(self, segment: Segment, file_name: str, pass_num: int) -> list[str]:
        """
        Return a complete ffmpeg command and arguments to perform the last pass for the specified
        segment and input file.
        """
        title = os.path.splitext(os.path.basename(file_name))[0]

        result = self.get_base_segment_args(segment, file_name)
        result += self.get_video_codec_args(segment, pass_num)
        result += self.get_audio_codec_args(segment)
        result += [
            '-pass', str(pass_num),
            '-passlogfile', title,
            '-metadata', f'title={title}'
        ]
        result += self.get_audio_metadata_map_args()
        result += self.args.passthrough_args

        safe_file_name = self.get_safe_filename(f'{title}{self.args.container.extension}')
        result.append(safe_file_name)

        return result

    # --------------------------------------------------------------------------------------------------
    def get_log_command(self, file_name: str) -> list[str]:
        """
        Return a list of arguments to either delete or rename the pass one log file, as requested by the
        user in the script arguemnts.
        """
        title = os.path.splitext(os.path.basename(file_name))[0]
        if self.args.delete_log:
            return ['rm', f'{title}-0.log']
        else:
            return ['mv', f'{title}-0.log', f'{title}_{datetime.now():%Y%m%d-%H%M%S}.log']

    # --------------------------------------------------------------------------------------------------
    def get_segment_commands(self, segment: Segment, file_name: str) -> None:
        """
        Return the list of commands to transcode the specified segment of the specified input file.
        """
        if self.video_format.passes == 1:
            return [self.get_one_pass_command(segment, file_name)]

        if 'pass_num' not in self.args or self.args.pass_num is None:
            passes = range(1, self.video_format.passes + 1)
        else:
            passes = [ int(self.args.pass_num) ]

        commands = []
        for pass_num in passes:
            if pass_num < self.video_format.passes:
                commands.append(self.get_not_last_pass_command(segment, file_name, pass_num))
            else:
                commands.append(self.get_last_pass_command(segment, file_name, pass_num))
                commands.append(self.get_log_command(file_name))

        return commands
